import logging

import stripe
from django.apps import apps
from django.conf import settings
from django.utils import timezone
from decimal import Decimal as D, ROUND_HALF_UP


logger = logging.getLogger(__name__)
Source = apps.get_model('payment', 'Source')
Order = apps.get_model('order', 'Order')

# https://support.stripe.com/questions/which-zero-decimal-currencies-does-stripe-support
ZERO_DECIMAL_CURRENCIES = (
    'BIF',  # Burundian Franc
    'CLP',  # Chilean Peso
    'DJF',  # Djiboutian Franc
    'GNF',  # Guinean Franc
    'JPY',  # Japanese Yen
    'KMF',  # Comorian Franc
    'KRW',  # South Korean Won
    'MGA',  # Malagasy Ariary
    'PYG',  # Paraguayan Guaraní
    'RWF',  # Rwandan Franc
    'VND',  # Vietnamese Đồng
    'VUV',  # Vanuatu Vatu
    'XAF',  # Central African Cfa Franc
    'XOF',  # West African Cfa Franc
    'XPF',  # Cfp Franc
)


class PaymentItem:
    def __init__(self, **kwargs):
        self.quantity = kwargs.get('quantity')
        self.title = kwargs.get('title')
        self.price_incl_tax = kwargs.get('price_incl_tax')
        self.price_currency = kwargs.get('price_currency')

class Facade(object):
    def __init__(self):
        stripe.api_key = settings.STRIPE_SECRET_KEY
        stripe.api_version = "2020-03-02"

    @staticmethod
    def get_friendly_decline_message(error):
        return 'The transaction was declined by your bank - please check your bankcard details and try again'

    @staticmethod
    def get_friendly_error_message(error):
        return 'An error occurred when communicating with the payment gateway.'

    def convert_to_cents(self, price, currency):
        """
        Convert price to cents with proper rounding, handling zero-decimal currencies.
        """
        if currency.upper() in ZERO_DECIMAL_CURRENCIES:
            return int(D(str(price)).quantize(D('1'), ROUND_HALF_UP))
        else:
            return int(D(str(price)).quantize(D('0.01'), ROUND_HALF_UP) * 100)

    def begin(self, customer_email, basket, total, shipping_method):
        metadata = {
            "discounts": "",
        }

        for voucher in basket.grouped_voucher_discounts:
            metadata['discounts'] += "{0} ({1}), ".format(voucher['voucher'].name, voucher['discount'])

        all_line_items = []
        for line in basket.all_lines():
            # this loop will split line into discounted and non-discounted lines
            for prices in line.get_price_breakdown():
                price_incl_tax, _, quantity = prices
                all_line_items.append(
                    PaymentItem(
                        quantity=quantity,
                        title=line.product.title,
                        price_incl_tax=price_incl_tax,
                        price_currency=line.price_currency
                    )
                )

        if basket.is_shipping_required() and shipping_method:
            price = shipping_method.calculate(basket)
            all_line_items.append(
                PaymentItem(
                    quantity=1,
                    title=shipping_method.name,
                    price_incl_tax=price.incl_tax,
                    price_currency=price.currency
                )
            )

        # Four cases here.  Two versions of the Stripe API, and whether we want to compress the line items or not.
        if settings.STRIPE_COMPRESS_TO_ONE_LINE_ITEM:
            line_items_summary = ", ".join(["{0}x{1}".format(l.quantity, l.title) for l in all_line_items])
            if not settings.STRIPE_USE_PRICES_API:
                line_items = [{
                    "name": line_items_summary,
                    "amount": self.convert_to_cents(total.incl_tax, total.currency),
                    "currency": total.currency,
                    "quantity": 1,
                }]
            else:
                line_items = [{
                    "price_data":  {
                        "product_data": {
                            "name": line_items_summary,
                        },
                        "currency": total.currency,
                        "unit_amount": self.convert_to_cents(total.incl_tax, total.currency),
                    },
                    "quantity": 1,
                }]
        else:
            line_items = []
            if not settings.STRIPE_USE_PRICES_API:
                for line_item in all_line_items:
                    line_items.append({
                        "name": line_item.title,
                        "amount": self.convert_to_cents(line_item.price_incl_tax, line_item.price_currency),
                        "currency": line_item.price_currency,
                        "quantity": line_item.quantity
                    })
            else:
                for line_item in all_line_items:
                    line_items.append({
                        "price_data":  {
                            "product_data": {
                                "name": line_item.title,
                            },
                            "currency": line_item.price_currency,
                            "unit_amount": self.convert_to_cents(line_item.price_incl_tax, line_item.price_currency),
                        },
                        "quantity": line_item.quantity,
                    })

        basket.freeze()
        session = stripe.checkout.Session.create(
            mode="payment",
            customer_email=customer_email,
            payment_method_types=['card'],
            line_items=line_items,
            metadata=metadata,
            success_url=settings.STRIPE_PAYMENT_SUCCESS_URL.format(basket.id),
            cancel_url=settings.STRIPE_PAYMENT_CANCEL_URL.format(basket.id),
            payment_intent_data={
                'capture_method': 'manual',
            },
        )
        return session

    def retrieve_payment_intent(self, pi):
        return stripe.PaymentIntent.retrieve(pi)

    def capture(self, order_number, **kwargs):
        """
        if capture is set to false in charge, the charge will only be pre-authorized
        one need to use capture to actually charge the customer
        """
        logger.info("Initiating payment capture for order '%s' via stripe" % (order_number))
        try:
            order = Order.objects.get(number=order_number)
            payment_source = Source.objects.get(order=order)
            # get charge_id from source
            charge_id = payment_source.reference

            stripe.PaymentIntent.modify(
                charge_id,
                receipt_email=order.user.email
            )

            stripe.PaymentIntent.capture(charge_id)
            # set captured timestamp
            payment_source.date_captured = timezone.now()
            payment_source.save()
            logger.info("payment for order '%s' (id:%s) was captured via stripe (stripe_ref:%s)" % (order.number, order.id, charge_id))
        except Source.DoesNotExist as e:
            logger.exception('Source Error for order: \'{}\''.format(order_number) )
            raise Exception("Capture Failure could not find payment source for Order %s" % order_number)
        except Order.DoesNotExist as e:
            logger.exception('Order Error for order: \'{}\''.format(order_number) )
            raise Exception("Capture Failure Order %s does not exist" % order_number)



from oscar.apps.checkout.apps import CheckoutConfig
from oscar.core.loading import get_class
from django.urls import path


class StripeSCACheckoutConfig(CheckoutConfig):
    def ready(self):
        stripe_payment_details_view = get_class("oscar_stripe_sca.views", "StripeSCAPaymentDetailsView")
        stripe_success_view = get_class("oscar_stripe_sca.views", "StripeSCASuccessResponseView")
        stripe_cancel_view = get_class("oscar_stripe_sca.views", "StripeSCACancelResponseView")
        super().ready()

    def get_urls(self):
        urls = super(StripeSCACheckoutConfig, self).get_urls()
        urls += [
            path('payment-details-stripe/',
                self.stripe_payment_details_view.as_view(), name='stripe-payment-details'),
            path('preview-stripe/<int:basket_id>/',
                self.stripe_success_view.as_view(preview=True), name='stripe-preview'),
            path('payment-cancel/<int:basket_id>/',
                self.stripe_cancel_view.as_view(), name='stripe-cancel'),
        ]
        return urls
    



from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.contrib import messages
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import RedirectView
from oscar.apps.checkout.views import PaymentDetailsView as CorePaymentDetailsView
from oscar.core.exceptions import ModuleNotFoundError
from oscar.core.loading import get_class, get_model
from oscar_stripe_sca.facade import logger
from django.utils.translation import gettext_lazy as _
from django.contrib import messages
from oscar_stripe_sca.facade import Facade
from . import PAYMENT_METHOD_STRIPE, PAYMENT_EVENT_PURCHASE

SourceType = get_model('payment', 'SourceType')
Source = get_model('payment', 'Source')
Line = get_model('basket', 'Line')
Basket = get_model('basket', 'Basket')
Selector = get_class('partner.strategy', 'Selector')
try:
    Applicator = get_class('offer.applicator', 'Applicator')
except ModuleNotFoundError:
    # fallback for django-oscar<=1.1
    Applicator = get_class('offer.utils', 'Applicator')


class StripeSCAPaymentDetailsView(CorePaymentDetailsView):
    template_name = "oscar_stripe_sca/stripe_payment_details.html"

    def get_context_data(self, **kwargs):
        ctx = super(StripeSCAPaymentDetailsView, self).get_context_data(**kwargs)
        customer_email = None
        try:
            customer_email = ctx["basket"].owner.email
        except AttributeError:
            checkout_data = self.request.session[self.checkout_session.SESSION_KEY]
            customer_email = checkout_data["guest"]["email"]
        stripe_session = Facade().begin(
            customer_email,
            ctx["basket"],
            ctx["order_total"],
            ctx["shipping_method"])
        self.request.session["stripe_session_id"] = stripe_session.id
        self.request.session["stripe_payment_intent_id"] = stripe_session.payment_intent
        ctx['stripe_publishable_key'] = settings.STRIPE_PUBLISHABLE_KEY
        ctx['stripe_session_id'] = stripe_session.id
        return ctx


class StripeSCASuccessResponseView(CorePaymentDetailsView):
    preview = True
    template_name_preview = 'oscar_stripe_sca/stripe_preview.html'

    @property
    def pre_conditions(self):
        return []

    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        return super(StripeSCASuccessResponseView, self).dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        ctx = super(StripeSCASuccessResponseView, self).get_context_data(**kwargs)
        if ctx['order_total'] is None:
            messages.error(self.request, "Your checkout session has expired, please try again")
            raise PermissionDenied
        else:
            ctx['order_total_incl_tax_cents'] = (
                ctx['order_total'].incl_tax * 100
            ).to_integral_value()
        return ctx

    def handle_payment(self, order_number, order_total, **kwargs):
        pi = self.request.session["stripe_payment_intent_id"]
        intent = Facade().retrieve_payment_intent(pi)
        intent.capture()

        source_type, __ = SourceType.objects.get_or_create(name=PAYMENT_METHOD_STRIPE)
        source = Source(
            source_type=source_type,
            currency=order_total.currency,
            amount_allocated=order_total.incl_tax,
            amount_debited=order_total.incl_tax,
            reference=pi)
        self.add_payment_source(source)

        self.add_payment_event(PAYMENT_EVENT_PURCHASE, order_total.incl_tax, reference=pi)

        del self.request.session["stripe_session_id"]
        del self.request.session["stripe_payment_intent_id"]

    def payment_description(self, order_number, total, **kwargs):
        return "Stripe payment for order {0} by {1}".format(order_number, self.request.user.get_full_name())

    @staticmethod
    def payment_metadata(order_number, total, **kwargs):
        return {
            'order_number': order_number,
        }

    def load_frozen_basket(self, basket_id):
        # Lookup the frozen basket that this txn corresponds to
        try:
            basket = Basket.objects.get(id=basket_id, status=Basket.FROZEN)
        except Basket.DoesNotExist:
            return None

        # Assign strategy to basket instance
        if Selector:
            basket.strategy = Selector().strategy(self.request)

        # Re-apply any offers
        Applicator().apply(basket, self.request.user, request=self.request)

        return basket

    def get(self, request, *args, **kwargs):
        kwargs['basket'] = self.load_frozen_basket(kwargs['basket_id'])
        if not kwargs['basket']:
            logger.warning(
                "Unable to load frozen basket with ID %s", kwargs['basket_id'])
            messages.error(
                self.request,
                _("No basket was found that corresponds to your "
                  "Stripe transaction"))
            return HttpResponseRedirect(reverse('basket:summary'))
        return super(StripeSCASuccessResponseView, self).get(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        """
        Place an order.
        """
        # Reload frozen basket which is specified in the URL
        basket = self.load_frozen_basket(kwargs['basket_id'])
        if not basket:
            messages.error(self.request, _("No basket was found that corresponds to your "
                  "Stripe transaction"))
            return HttpResponseRedirect(reverse('basket:summary'))

        submission = self.build_submission(basket=basket)
        return self.submit(**submission)


class StripeSCACancelResponseView(RedirectView):
    permanent = False

    def get(self, request, *args, **kwargs):
        basket = get_object_or_404(Basket, id=kwargs['basket_id'],
                                   status=Basket.FROZEN)
        basket.thaw()
        logger.info("Payment cancelled (token %s) - basket #%s thawed",
                    request.GET.get('token', '<no token>'), basket.id)
        return super(StripeSCACancelResponseView, self).get(request, *args, **kwargs)

    def get_redirect_url(self, **kwargs):
        messages.error(self.request, _("Stripe transaction cancelled"))
        return reverse('basket:summary')



from django.conf import settings
from django.urls import reverse_lazy

STRIPE_SEND_RECEIPT = getattr(settings, "STRIPE_SEND_RECEIPT", True)
STRIPE_PUBLISHABLE_KEY = getattr(settings, "STRIPE_PUBLISHABLE_KEY", None)
STRIPE_SECRET_KEY = getattr(settings, "STRIPE_SECRET_KEY", None)
STRIPE_COMPRESS_TO_ONE_LINE_ITEM = getattr(settings, "STRIPE_COMPRESS_TO_ONE_LINE_ITEM", True)
STRIPE_USE_PRICES_API = getattr(settings, "STRIPE_USE_PRICES_API", True)
STRIPE_RETURN_URL_BASE = getattr(settings, "STRIPE_RETURN_URL_BASE", "http://localhost/")
STRIPE_PAYMENT_SUCCESS_URL = getattr(settings, "STRIPE_PAYMENT_SUCCESS_URL", "{0}{1}".format(settings.STRIPE_RETURN_URL_BASE, reverse_lazy("checkout:stripe-preview")))
STRIPE_PAYMENT_CANCEL_URL = getattr(settings, "STRIPE_PAYMENT_CANCEL_URL", "{0}{1}".format(settings.STRIPE_RETURN_URL_BASE, reverse_lazy("checkout:stripe-cancel")))